import requests
import sys
import json
import base64


class REST:
    def __init__(self, key, domain, base_url="https://chat-api.imi.chat"):
        self.key = key
        self.engage_domain = domain
        self.auth_code = base64.b64encode(f"{key}@{domain}".encode("ascii")).decode('UTF-8')
        self.base_url = base_url
        test_auth = self.authenticate()

    def authenticate(self) -> str:
        try:
            token = requests.get(f"{self.base_url}/authorize",
                                 headers={"auth_code": self.auth_code,
                                          "Content-Type": "application/json"}
                                 ).headers['access_token']
        except Exception as err:
            raise Exception(f"Authentication Error: {str(err)}")
        else:
            return token

    def get(self, uri: str, params: dict = {}, headers: dict = {}) -> dict:
        uri = f"/{uri}" if uri.find('/') != 0 else uri
        if uri.find(self.base_url) > 0:
            uri = uri
        else:
            uri = f"{self.base_url}{uri}"

        headers.update({"access_token": self.authenticate(),
                        "Content-Type": "applicaton/json"})

        try:
            response = requests.get(f"{uri}",
                                    headers=headers,
                                    data=params
                                    ).json()
        except Exception as err:
            print(f"GET ERROR: {uri} - {err}", file=sys.stderr)
        else:
            if response['response'] is None:
                print(f"REST GET Error: {uri} - {response['status_code']} - {response['description']}",
                      file=sys.stderr)
                return {}
            return response['response']

    def post(self, uri: str, params: dict = {}, headers: dict = {}) -> dict:
        uri = f"/{uri}" if uri.find('/') != 0 else uri
        if uri.find(self.base_url) > 0:
            uri = uri
        else:
            uri = f"{self.base_url}{uri}"

        headers.update({"access_token": self.authenticate()})
        # "Content-Type": "applicaton/json"})

        try:
            response = requests.post(f"{uri}",
                                     headers=headers,
                                     json=params
                                     ).json()
        except Exception as err:
            print(f"POST ERROR: {uri} - {err}", file=sys.stderr)
        else:
            if 'response' not in response.keys() or response['response'] is None:
                print(f"REST POST Error: {uri} - {response} - {params}",
                      file=sys.stderr)
                return {}
            return response['response']

    def patch(self, uri: str, params: dict = {}, headers: dict = {}) -> dict:
        uri = f"/{uri}" if uri.find('/') != 0 else uri
        if uri.find(self.base_url) > 0:
            uri = uri
        else:
            uri = f"{self.base_url}{uri}"

        headers.update({"access_token": self.authenticate()})
        # "Content-Type": "applicaton/json"})

        try:
            response = requests.patch(f"{uri}",
                                      headers=headers,
                                      json=params
                                      ).json()
        except Exception as err:
            print(f"PATCH ERROR: {uri} - {err}", file=sys.stderr)
        else:
            if 'response' not in response.keys() or response['response'] is None:
                print(f"REST PATCH Error: {uri} - {response} - {params}",
                      file=sys.stderr)
                return {}
            return response['response']

    def put(self, uri: str, params: dict = {}, headers: dict = {}) -> dict:
        uri = f"/{uri}" if uri.find('/') != 0 else uri
        if uri.find(self.base_url) > 0:
            uri = uri
        else:
            uri = f"{self.base_url}{uri}"

        headers.update({"access_token": self.authenticate()})
        # "Content-Type": "applicaton/json"})

        try:
            response = requests.put(f"{uri}",
                                    headers=headers,
                                    json=params
                                    ).json()
        except Exception as err:
            print(f"PUT ERROR: {uri} - {err}", file=sys.stderr)
        else:
            if 'response' not in response.keys() or response['response'] is None:
                print(f"REST PUT Error: {uri} - {response} - {params}",
                      file=sys.stderr)
                return {}
            return response['response']



from zeep.transports import Transport
from zeep import Client, Settings
from requests import Session
from requests.auth import HTTPBasicAuth
import sys


class Logs():
    class ProgressBar():
        progress_bar_scale = .7
        progress_bar_length = int(100 * progress_bar_scale)
        completed_bar_char = "█"
        incomplete_bar_char = "▒"
        completion_message = "Completed!"

        def __init__(self, total):
            print()
            self.total = total

        def progress(self, current):
            # print out our progress bar
            percent_complete = round(current / self.total * 100)
            bars_of_completion = int(percent_complete * self.progress_bar_scale)
            bars_to_be_completed = self.progress_bar_length - bars_of_completion
            print(
                f'\r{self.completed_bar_char * bars_of_completion}{self.incomplete_bar_char * bars_to_be_completed} {percent_complete}%',
                end='')

    def __init__(self,ipaddr,username,passwd):
        # set WSDL file from URL
        wsdl = f'https://{ipaddr}:8443/logcollectionservice2/services/LogCollectionPortTypeService?wsdl'

        # create a SOAP client session
        session = Session()

        # avoid certificate verification by default and setup session
        session.verify = False
        session.auth = HTTPBasicAuth(username, passwd)
        transport = Transport(session=session, timeout=10)
        settings = Settings(strict=False, xml_huge_tree=True)

        # Create client and soap service
        client = Client(wsdl, settings=settings,
                                transport=transport)  # , plugins = plugin )
        self.client = client.create_service(
            '{http://cisco.com/ccm/serviceability/soap/LogCollection/GetFile/}GetFileBinding',
            wsdl)

    # File request from CM via DIME
    def get_file(self, filepath):
        # check if file path includes /var/log as required if not assume /var/log/active
        if filepath.find('/var/log') == -1:
            filepath = f"/var/log/active/{filepath}"

        try:
            file_contents = self.client.GetOneFile(filepath).attachments[0].content.decode('UTF-8')

        except Exception as err:
            print(f"Error pulling {filepath} from CUCM - {err}", file=sys.stderr)
            return False

        else:
            return file_contents



from .REST import REST
from .User import User
from .Team import Team
from .Chat import Chat


class Connect:
    def __init__(self, key=None, domain=None):
        self.rest = REST(key, domain)
        self.user = User(self)
        self.team = Team(self)
        self.chat = Chat(self)



import requests
from requests.auth import HTTPBasicAuth
import sys
import xmltodict
import hashlib
import urllib

requests.packages.urllib3.disable_warnings()
try:
    requests.packages.urllib3.contrib.pyopenssl.util.ssl_.DEFAULT_CIPHERS += ':HIGH:!DH:!aNULL'
    requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS += ':HIGH:!DH:!aNULL'
except AttributeError:
    # no pyopenssl support used / needed / available
    pass

REQUIRED_LIBS = {}


class Connect:

    # initialize object set system type and build DB connectors as needed
    def __init__(self,
                 ipaddr=None,
                 username=None,
                 passwd=None):

        # if type= cucm then set username/password for AXL connection
        if ipaddr is  None or passwd is None or username is None:
            print(f'Usage: CollabConnector.CUCM("ipaddr", "admin", "password", wsdl="./AXLAPI.wsdl")',
                  file=sys.stderr)
        else:
            self.system_type = "CER"
            self.ipaddr = ipaddr
            self.username = username
            self.auth = HTTPBasicAuth(username, passwd)
            self.hash_pass = hashlib.sha256(passwd.encode()).hexdigest()

            print("Connected.")

    # REST Wrapper for CER
    def cer_api(self, target_uri, method='GET', data={}):
        if target_uri.find("cerappservices/export/") > -1:
            target_uri = target_uri
        else:
            if target_uri.find('/') == 0:
                target_uri = target_uri[1:]
            target_uri = f"https://{self.ipaddr}:8443/cerappservices/export/{target_uri}"

        if len(data) > 0:
            target_uri += f"?{urllib.parse.urlencode(data)}"

        target_uri += f"/{self.username}/{self.hash_pass}"

        # create result array
        try:
            response = requests.request(method, target_uri, verify=False, timeout=(5, 30))
        except Exception as err:
            print(f"\tError with REST call - {err}")
            return False

        if str(response.status_code) == 200:
            try:
                cer_array = xmltodict.parse(response.text, dict_constructor=dict)
            except Exception as err:
                print("Error parsing XML.  Expect String back", file=sys.stderr)
                return response.text
            else:
                return cer_array

        else:
            print("\tError:" + str(response) + "\n\n" + response.text, file=sys.stderr)
            return False

    # wrapper for api Gets
    def get(self, target_endpoint, params={}):
        return self.cer_api(target_endpoint, data=params)



import sys


class Team:
    def __init__(self, parent):
        self.parent = parent

    def list(self):
        if teams := self.parent.rest.get(f"/v3.0/teams"):
            return teams['teams']
        return teams

    def find(self, name: str) -> dict:
        teams = self.parent.rest.get(f"/v3.0/teams")
        if teams and 'teams' in teams.keys():
            teams = teams['teams']
        if name is None:
            return teams

        for team in self.parent.rest.get(f"/v3.0/teams")['teams']:
            if name.lower() == team['name'].lower():
                return team
        print(f"Team not found: {name}", file=sys.stderr)
        return {}

    def get(self, team_id: int) -> dict:
        if team := self.parent.rest.get(f"/v3.0/teams/{team_id}"):
            return team

        print(f"Error getting team: {team_id}", file=sys.stderr)
        return {}



import requests
import xmltodict
from requests.auth import HTTPBasicAuth
import sys
import re
import json
import urllib
import socket

requests.packages.urllib3.disable_warnings()
try:
    requests.packages.urllib3.contrib.pyopenssl.util.ssl_.DEFAULT_CIPHERS += ':HIGH:!DH:!aNULL'
    requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS += ':HIGH:!DH:!aNULL'
except AttributeError:
    # no pyopenssl support used / needed / available
    pass

class Connect:

    # initialize object set system type and build DB connectors as needed
    def __init__(self,
                 ipaddr=None,
                 username=None,
                 passwd=None,
                 db=False,
                 server_name=None
                 ):

        self.system_type = "UCCX"
        self.ipaddr = ipaddr

        # if type= cucm then set username/password for AXL connection
        if ipaddr is None or passwd is None or username is None:
            raise Exception(f'Usage: CollabConnector.UCCX.Connect("ipaddr", "admin", "password", db=False)')
        else:
            self.username = username
            self.auth = HTTPBasicAuth(username, passwd)

            if self.open_port(ipaddr, 8443) is False:
                raise Exception(f"Connection Error: {ipaddr}:8443 not reachable or open. Is this UCCX?")
            if self.get("applications") is False:
                raise Exception(f"Connection Error: API request not valid. Improper credentials?")

            # if type = uccx-db then set the informixConnection variable for the connector
            if db:
                try:
                    import IfxPy
                except:
                    print("""
                            IfxPy module failed to load. If you intend to connect to an Informix DB check the linux enviroment
                            << pip3 install IfxPy >>
                            export INFORMIXDIR=/opt/IBM/Informix_Client-SDK/  
                            export CSDK_HOME=$INFORMIXDIR
                            export LIBPATH=${INFORMIXDIR}/lib:${INFORMIXDIR}/lib/cli:${INFORMIXDIR}/lib/esql:${INFORMIXDIR}/lib:${INFORMIXDIR}/bin:${INFORMIXDIR}/etc:${LIBPATH}
                            export LD_LIBRARY_PATH=$INFORMIXDIR/lib:$INFORMIXDIR/lib/cli:$INFORMIXDIR/lib/esql
                            """, file=sys.stderr)
                    self.INFORMIX = False
                else:
                    connection_string = f"SERVER={server_name};DATABASE=db_cra;HOST={ipaddr};SERVICE=5104;UID=uccxhruser;PWD={passwd};DB_LOCALE=en_US.utf8;"
                    try:
                        self.informixConnector = IfxPy.connect(connection_string, "", "")
                    except Exception as err:
                        print(f'INFORMIX connection failed. {err}', file=sys.stderr)
                        self.INFORMIX = False
                    else:
                        self.INFORMIX = True
            else:
                self.INFORMIX = False
                print("Informix Drivers not correctly initialized.  Try again or use the API", file=sys.stderr)

    @staticmethod
    def open_port(ip, port, return_object=[]):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(3)
        try:
            s.connect((ip, int(port)))
            s.close()
            return_object.append(str(port))
            return port
        except Exception as e:
            s.close()
            return False

    # Function query SQL via direct Informix connector
    def informix_query(self, sql_statement):
        if self.INFORMIX is False:
            print("Not connected to INFORMIX DB. Use db=True.", file=sys.stderr)
            return False

        result = []
        if re.search("select", sql_statement.lower()) and not re.search('select.*limit .*from', sql_statement.lower()):
            selectCount = re.sub("[sS][eE][lL][eE][cC][tT] .* [fF][rR][oO][mM] ", "SELECT COUNT(*) AS qty FROM ",
                                 sql_statement)
            stmt = IfxPy.exec_immediate(self.informixConnector, selectCount)
            return_rows = int(IfxPy.fetch_assoc(stmt)['qty'])

            request_count = 0
            while request_count < return_rows:
                request_rows = re.sub("[sS][eE][lL][eE][cC][tT] ", f"SELECT SKIP {request_count} LIMIT 4000 ",
                                      sql_statement)

                try:
                    stmt = IfxPy.exec_immediate(self.informixConnector, request_rows)
                except Exception as err:
                    print(f"SQL Error: {err}", file=sys.stderr)
                    return False
                else:
                    while True:
                        assoc = IfxPy.fetch_assoc(stmt)
                        if assoc is not False:
                            result.append(assoc)
                        else:
                            break

                request_count = len(result)

        else:
            try:
                stmt = IfxPy.exec_immediate(self.informixConnector, sql_statement)
            except Exception as err:
                print(f"SQL Error: {err}", file=sys.stderr)
                return False
            else:
                assoc = IfxPy

            while True:
                assoc = IfxPy.fetch_assoc(stmt)
                if assoc is not False:
                    result.append(assoc)
                else:
                    break

        return result

    # REST Wrapper for UCCX
    def uccx_api(self, target_uri, method='GET', data={}):
        if target_uri.find("adminapi/") > -1:
            target_uri = "".join(target_uri.split("adminapi/")[1:])
        target_uri = f"https://{self.ipaddr}:8443/adminapi/{target_uri}"

        if len(data) > 0:
            target_uri += f"?{urllib.parse.urlencode(data)}"
        print(target_uri)
        attempt = 0
        while attempt < 2:
            try:
                # send API request
                response = requests.request(method, target_uri, auth=self.auth,
                                            headers={'Accept': 'application/json', 'Content-Type': 'application/json'},
                                            data=json.dumps(data), verify=False)
            except Exception as err:
                print(f"Error requesting API: {err}", file=sys.stderr)
                attempt += 1
            else:
                # format response into result array
                if 200 <= response.status_code <= 300:
                    try:
                        result = json.loads(response.text)
                    except Exception as err:
                        return [response.text]
                    else:
                        if "apiError" in result.keys():
                            print(f"Error requesting CCX API: {target_uri} - {response.text}", file=sys.stderr)
                            return False

                        elif isinstance(result, dict):
                            if target_uri in result:
                                if isinstance(result[target_uri], list):
                                    return result[target_uri]
                                else:
                                    return [result[target_uri]]
                            else:
                                if isinstance(result, list):
                                    return result
                                else:
                                    return [result]
                        else:
                            return [result]
                elif 400 <= response.status_code <= 600:
                    print(f"Error requesting CCX API: {target_uri} - {response.text}", file=sys.stderr)
                    return False
                else:
                    return response.text

        return False

    # wrapper for api Gets
    def get(self, target_endpoint, params={}):
        return self.uccx_api(target_endpoint, data=params)

    # wrapper for api PUT
    def put(self, target_endpoint, put_data):
        return self.uccx_api(target_endpoint, method='PUT', data=put_data)

    # wrapper for api POST
    def post(self, target_endpoint, post_data):
        return self.uccx_api(target_endpoint, method='POST', data=post_data)

    # wrapper for api DELETE
    def delete(self, target_endpoint):
        return self.uccx_api(target_endpoint, method='DELETE')

    # wrapper for api PATCH
    def patch(self, target_endpoint, patch_data):
        return self.uccx_api(target_endpoint, method='PATCH', data=patch_data)



import sqlite3
import sys
from .CDROnDemand import *
try:
    import pandas as pd
except:
    print("Error importing Pandas for CDR", file=sys.stderr)
    print("<< pip3 install pandas >>", file=sys.stderr)


class CDR:
    _TABLE_NAME = None
    on_demand = None

    def __init__(self, ipaddr=None, username=None, passwd=None):
        if ipaddr and username and passwd:
            self.on_demand = CDROnDemand(ipaddr, username, passwd)

    def parse(self, db=None, csv=None, df=None, cdr=None, cmr=None, table="cdr_cdr", cmr_table="cdr_cmr"):
        if csv is None and cdr is not None:
            csv = cdr
        if csv is None and db is None and df is None:
            print("You must specify an Existing DB or a CSV or DataFrame object to import", file=sys.stderr)

        else:
            self._TABLE_NAME = table
            self._TABLE_CMR = cmr_table

            if db is not None:
                self._DB_NAME = db  # set name of DB File if not given
            else:
                if csv is not None:
                    self._DB_NAME = f"{csv}.sqlite"
                elif df is not None:
                    self._DB_NAME = f"DataFramesImport.sqlite"

            if csv is not None or df is not None:
                start = time.time()

                # convert CSV to Pandas DataFrame
                if csv is not None:
                    try:
                        dtypes = {"cdrRecordType": int,
                                  "globalCallID_callManagerId": int,
                                  "globalCallID_callId": int,
                                  "origLegCallIdentifier": int,
                                  "dateTimeOrigination": int,
                                  "origNodeId": int,
                                  "origSpan": int,
                                  "origIpAddr": int,
                                  "callingPartyNumber": str,
                                  "callingPartyUnicodeLoginUserID": str,
                                  "origCause_location": int,
                                  "origCause_value": int,
                                  "origPrecedenceLevel": int,
                                  "origMediaTransportAddress_IP": int,
                                  "origMediaTransportAddress_Port": int,
                                  "origMediaCap_payloadCapability": int,
                                  "origMediaCap_maxFramesPerPacket": int,
                                  "origMediaCap_g723BitRate": int,
                                  "origVideoCap_Codec": int,
                                  "origVideoCap_Bandwidth": int,
                                  "origVideoCap_Resolution": int,
                                  "origVideoTransportAddress_IP": int,
                                  "origVideoTransportAddress_Port": int,
                                  "origRSVPAudioStat": str,
                                  "origRSVPVideoStat": str,
                                  "destLegIdentifier": int,
                                  "destNodeId": int,
                                  "destSpan": int,
                                  "destIpAddr": int,
                                  "originalCalledPartyNumber": str,
                                  "finalCalledPartyNumber": str,
                                  "finalCalledPartyUnicodeLoginUserID": str,
                                  "destCause_location": int,
                                  "destCause_value": int,
                                  "destPrecedenceLevel": int,
                                  "destMediaTransportAddress_IP": int,
                                  "destMediaTransportAddress_Port": int,
                                  "destMediaCap_payloadCapability": int,
                                  "destMediaCap_maxFramesPerPacket": int,
                                  "destMediaCap_g723BitRate": int,
                                  "destVideoCap_Codec": int,
                                  "destVideoCap_Bandwidth": int,
                                  "destVideoCap_Resolution": int,
                                  "destVideoTransportAddress_IP": int,
                                  "destVideoTransportAddress_Port": int,
                                  "destRSVPAudioStat": str,
                                  "destRSVPVideoStat": str,
                                  "dateTimeConnect": int,
                                  "dateTimeDisconnect": int,
                                  "lastRedirectDn": str,
                                  "pkid": str,
                                  "originalCalledPartyNumberPartition": str,
                                  "callingPartyNumberPartition": str,
                                  "finalCalledPartyNumberPartition": str,
                                  "lastRedirectDnPartition": str,
                                  "duration": int,
                                  "origDeviceName": str,
                                  "destDeviceName": str,
                                  "origCallTerminationOnBehalfOf": int,
                                  "destCallTerminationOnBehalfOf": int,
                                  "origCalledPartyRedirectOnBehalfOf": int,
                                  "lastRedirectRedirectOnBehalfOf": int,
                                  "origCalledPartyRedirectReason": int,
                                  "lastRedirectRedirectReason": int,
                                  "destConversationId": int,
                                  "globalCallId_ClusterID": str,
                                  "joinOnBehalfOf": int,
                                  "comment": str,
                                  "authCodeDescription": str,
                                  "authorizationLevel": int,
                                  "clientMatterCode": str,
                                  "origDTMFMethod": int,
                                  "destDTMFMethod": int,
                                  "callSecuredStatus": int,
                                  "origConversationId": int,
                                  "origMediaCap_Bandwidth": int,
                                  "destMediaCap_Bandwidth": int,
                                  "authorizationCodeValue": str,
                                  "outpulsedCallingPartyNumber": str,
                                  "outpulsedCalledPartyNumber": str,
                                  "origIpv4v6Addr": str,
                                  "destIpv4v6Addr": str,
                                  "origVideoCap_Codec_Channel2": int,
                                  "origVideoCap_Bandwidth_Channel2": int,
                                  "origVideoCap_Resolution_Channel2": int,
                                  "origVideoTransportAddress_IP_Channel2": int,
                                  "origVideoTransportAddress_Port_Channel2": int,
                                  "origVideoChannel_Role_Channel2": int,
                                  "destVideoCap_Codec_Channel2": int,
                                  "destVideoCap_Bandwidth_Channel2": int,
                                  "destVideoCap_Resolution_Channel2": int,
                                  "destVideoTransportAddress_IP_Channel2": int,
                                  "destVideoTransportAddress_Port_Channel2": int,
                                  "destVideoChannel_Role_Channel2": int,
                                  "incomingProtocolID": int,
                                  "incomingProtocolCallRef": str,
                                  "outgoingProtocolID": int,
                                  "outgoingProtocolCallRef": str,
                                  "currentRoutingReason": int,
                                  "origRoutingReason": int,
                                  "lastRedirectingRoutingReason": int,
                                  "huntPilotDN": str,
                                  "huntPilotPartition": str,
                                  "calledPartyPatternUsage": int,
                                  "outpulsedOriginalCalledPartyNumber": str,
                                  "outpulsedLastRedirectingNumber": str,
                                  "wasCallQueued": int,
                                  "totalWaitTimeInQueue": int,
                                  "callingPartyNumber_uri": str,
                                  "originalCalledPartyNumber_uri": str,
                                  "finalCalledPartyNumber_uri": str,
                                  "lastRedirectDn_uri": str,
                                  "mobileCallingPartyNumber": str,
                                  "finalMobileCalledPartyNumber": str,
                                  "origMobileDeviceName": str,
                                  "destMobileDeviceName": str,
                                  "origMobileCallDuration": int,
                                  "destMobileCallDuration": int,
                                  "mobileCallType": int,
                                  "originalCalledPartyPattern": str,
                                  "finalCalledPartyPattern": str,
                                  "lastRedirectingPartyPattern": str,
                                  "huntPilotPattern": str}
                        df = pd.concat(pd.read_csv(csv, dtype=dtypes, iterator=True, chunksize=8000))

                    except Exception as err:
                        print(f"Error converting CDR CSV to DataFrame - {err}", file=sys.stderr)
                    else:
                        self.create_table()  # create and load table based on CSV
                        self.load_cdr_table(df)  # create and load table based on DataFrame
                        print(f"\t... CDR DB Created in {(time.time() - start)} sec.")

            if cmr is not None:
                start = time.time()
                try:
                    df = pd.concat(pd.read_csv(cmr, iterator=True, chunksize=8000))
                except Exception as err:
                    print(f"Error converting CMR CSV to DataFrame - {err}", file=sys.stderr)
                else:
                    self.load_cmr_table(df)  # create and load table based on DataFrame
                    print(f"\t... CMR DB Created in {(time.time() - start)} sec.")

    def change_table(self, table):
        self._TABLE_NAME = table

    def create_table(self):
        _cdr_db = sqlite3.connect(self._DB_NAME)
        sql_cdr = f"""CREATE TABLE {self._TABLE_NAME}(
					cdrRecordType INTEGER NULL,
					globalCallID_callManagerId INTEGER NULL,
					globalCallID_callId INTEGER NULL,
					origLegCallIdentifier INTEGER NULL,
					dateTimeOrigination INTEGER NULL,
					origNodeId INTEGER NULL,
					origSpan INTEGER NULL,
					origIpAddr INTEGER NULL,
					callingPartyNumber VARCHAR(50) NULL,
					callingPartyUnicodeLoginUserID VARCHAR(50) NULL,
					origCause_location INTEGER NULL,
					origCause_value INTEGER NULL,
					origPrecedenceLevel INTEGER NULL,
					origMediaTransportAddress_IP INTEGER NULL,
					origMediaTransportAddress_Port INTEGER NULL,
					origMediaCap_payloadCapability INTEGER NULL,
					origMediaCap_maxFramesPerPacket INTEGER NULL,
					origMediaCap_g723BitRate INTEGER NULL,
					origVideoCap_Codec INTEGER NULL,
					origVideoCap_Bandwidth INTEGER NULL,
					origVideoCap_Resolution INTEGER NULL,
					origVideoTransportAddress_IP INTEGER NULL,
					origVideoTransportAddress_Port INTEGER NULL,
					origRSVPAudioStat VARCHAR(64) NULL,
					origRSVPVideoStat VARCHAR(64) NULL,
					destLegIdentifier INTEGER NULL,
					destNodeId INTEGER NULL,
					destSpan INTEGER NULL,
					destIpAddr INTEGER NULL,
					originalCalledPartyNumber VARCHAR(50) NULL,
					finalCalledPartyNumber VARCHAR(50) NULL,
					finalCalledPartyUnicodeLoginUserID VARCHAR(50) NULL,
					destCause_location INTEGER NULL,
					destCause_value INTEGER NULL,
					destPrecedenceLevel INTEGER NULL,
					destMediaTransportAddress_IP INTEGER NULL,
					destMediaTransportAddress_Port INTEGER NULL,
					destMediaCap_payloadCapability INTEGER NULL,
					destMediaCap_maxFramesPerPacket INTEGER NULL,
					destMediaCap_g723BitRate INTEGER NULL,
					destVideoCap_Codec INTEGER NULL,
					destVideoCap_Bandwidth INTEGER NULL,
					destVideoCap_Resolution INTEGER NULL,
					destVideoTransportAddress_IP INTEGER NULL,
					destVideoTransportAddress_Port INTEGER NULL,
					destRSVPAudioStat VARCHAR(64) NULL,
					destRSVPVideoStat VARCHAR(64) NULL,
					dateTimeConnect INTEGER NULL,
					dateTimeDisconnect INTEGER NULL,
					lastRedirectDn VARCHAR(50) NULL,
					pkid varchar(64) NULL,
					originalCalledPartyNumberPartition VARCHAR(50) NULL,
					callingPartyNumberPartition VARCHAR(50) NULL,
					finalCalledPartyNumberPartition VARCHAR(50) NULL,
					lastRedirectDnPartition VARCHAR(50) NULL,
					duration INTEGER NULL,
					origDeviceName VARCHAR(129) NULL,
					destDeviceName VARCHAR(129) NULL,
					origCallTerminationOnBehalfOf INTEGER NULL,
					destCallTerminationOnBehalfOf INTEGER NULL,
					origCalledPartyRedirectOnBehalfOf INTEGER NULL,
					lastRedirectRedirectOnBehalfOf INTEGER NULL,
					origCalledPartyRedirectReason INTEGER NULL,
					lastRedirectRedirectReason INTEGER NULL,
					destConversationId INTEGER NULL,
					globalCallId_ClusterID VARCHAR(50) NULL,
					joinOnBehalfOf INTEGER NULL,
					comment VARCHAR(2048) NULL,
					authCodeDescription VARCHAR(50) NULL,
					authorizationLevel INTEGER NULL,
					clientMatterCode VARCHAR(32) NULL,
					origDTMFMethod INTEGER NULL,
					destDTMFMethod INTEGER NULL,
					callSecuredStatus INTEGER NULL,
					origConversationId INTEGER NULL,
					origMediaCap_Bandwidth INTEGER NULL,
					destMediaCap_Bandwidth INTEGER NULL,
					authorizationCodeValue VARCHAR(32) NULL,
					outpulsedCallingPartyNumber VARCHAR(50) NULL,
					outpulsedCalledPartyNumber VARCHAR(50) NULL,
					origIpv4v6Addr VARCHAR(64) NULL,
					destIpv4v6Addr VARCHAR(64) NULL,
					origVideoCap_Codec_Channel2 INTEGER NULL,
					origVideoCap_Bandwidth_Channel2 INTEGER NULL,
					origVideoCap_Resolution_Channel2 INTEGER NULL,
					origVideoTransportAddress_IP_Channel2 INTEGER NULL,
					origVideoTransportAddress_Port_Channel2 INTEGER NULL,
					origVideoChannel_Role_Channel2 INTEGER NULL,
					destVideoCap_Codec_Channel2 INTEGER NULL,
					destVideoCap_Bandwidth_Channel2 INTEGER NULL,
					destVideoCap_Resolution_Channel2 INTEGER NULL,
					destVideoTransportAddress_IP_Channel2 INTEGER NULL,
					destVideoTransportAddress_Port_Channel2 INTEGER NULL,
					destVideoChannel_Role_Channel2 INTEGER NULL,
					incomingProtocolID INTEGER NULL,
					incomingProtocolCallRef VARCHAR(32) NULL,
					outgoingProtocolID INTEGER NULL,
					outgoingProtocolCallRef VARCHAR(32) NULL,
					currentRoutingReason INTEGER NULL,
					origRoutingReason INTEGER NULL,
					IncomingICID VARCHAR(50) NULL,
					IncomingOrigIOI VARCHAR(50) NULL,
					IncomingTermIOI VARCHAR(50) NULL,
					OutgoingICID VARCHAR(50) NULL,
					OutgoingOrigIOI VARCHAR(50) NULL,
					OutgoingTermIOI VARCHAR(50) NULL,
					lastRedirectingRoutingReason INTEGER NULL,
					huntPilotDN VARCHAR(50) NULL,
					huntPilotPartition VARCHAR(50) NULL,
					calledPartyPatternUsage INTEGER NULL,
					outpulsedOriginalCalledPartyNumber VARCHAR(50) NULL,
					outpulsedLastRedirectingNumber VARCHAR(50) NULL,
					wasCallQueued INTEGER NULL,
					totalWaitTimeInQueue INTEGER NULL,
					callingPartyNumber_uri VARCHAR(254) NULL,
					originalCalledPartyNumber_uri VARCHAR(254) NULL,
					finalCalledPartyNumber_uri VARCHAR(254) NULL,
					lastRedirectDn_uri VARCHAR(254) NULL,
					mobileCallingPartyNumber VARCHAR(50) NULL,
					finalMobileCalledPartyNumber VARCHAR(50) NULL,
					origMobileDeviceName VARCHAR(50) NULL,
					destMobileDeviceName VARCHAR(50) NULL,
					origMobileCallDuration INTEGER NULL,
					destMobileCallDuration INTEGER NULL,
					mobileCallType INTEGER NULL,
					originalCalledPartyPattern VARCHAR(50) NULL,
					finalCalledPartyPattern VARCHAR(50) NULL,
					lastRedirectingPartyPattern VARCHAR(50) NULL,
					huntPilotPattern VARCHAR(50) NULL,
					origDeviceType VARCHAR(100) NULL,
					destDeviceType VARCHAR(100) NULL,
					origDeviceSessionID VARCHAR(128) NULL,
					destDeviceSessionID VARCHAR(128) NULL
					);"""
        sql_cmr = f"""CREATE TABLE {self._TABLE_CMR} (
						cdrRecordType INTEGER ,
						globalCallID_callManagerId INTEGER ,
						globalCallId_callId INTEGER ,
						nodeId INTEGER ,
						directoryNumber INTEGER ,
						callIdentifier INTEGER ,
						dateTimeStamp INTEGER ,
						numberPacketsSent INTEGER ,
						numberOctetsSent INTEGER ,
						numberPacketsReceived INTEGER ,
						numberOctetsReceived INTEGER ,
						numberPacketsLost INTEGER ,
						jitter INTEGER ,
						latency INTEGER ,
						pkid VARCHAR(128),
						directoryNumberPartition VARCHAR(128),
						globalCallId_ClusterId VARCHAR(128),
						deviceName VARCHAR(128),
						varVQMetrics VARCHAR(128),
						duration INTEGER ,
						videoContentType VARCHAR(128),
						videoDuration INTEGER ,
						numberVideoPacketsSent INTEGER ,
						numberVideoOctetsSent INTEGER ,
						numberVideoPacketsReceived INTEGER ,
						numberVideoOctetsReceived INTEGER ,
						numberVideoPacketsLost INTEGER ,
						videoAverageJitter INTEGER ,
						videoRoundTripTime INTEGER ,
						videoOneWayDelay VARCHAR(128),
						videoTransmissionMetrics VARCHAR(128),
						videoReceptionMetrics VARCHAR(128),
						videoContentType_channel2 VARCHAR(128),
						videoDuration_channel2 INTEGER ,
						numberVideoPacketsSent_channel2 INTEGER ,
						numberVideoOctetsSent_channel2 INTEGER ,
						numberVideoPacketsReceived_channel2 INTEGER ,
						numberVideoOctetsReceived_channel2 INTEGER ,
						numberVideoPacketsLost_channel2 INTEGER ,
						videoAverageJitter_channel2 INTEGER ,
						videoRoundTripTime_channel2 INTEGER ,
						videoOneWayDelay_channel2 INTEGER ,
						videoReceptionMetrics_channel2 VARCHAR(128),
						videoTransmissionMetrics_channel2 VARCHAR(128),

						);"""
        # Build SQLite Table
        print("Building SQLite DB from CSV...")
        try:
            _cdr_db.execute(sql_cdr)

        except Exception as err:
            print(err, file=sys.stderr)

        try:
            _cdr_db.execute(sql_cmr)

        except Exception as err:
            print(err, file=sys.stderr)

        _cdr_db.commit()
        _cdr_db.close()
        return True

    def load_cdr_table(self, cdr_df):
        _cdr_db = sqlite3.connect(self._DB_NAME)

        print("Loading CDR Table")
        try:
            cdr_df.to_sql(self._TABLE_NAME, _cdr_db, if_exists='append', index=False)

        except Exception as err:
            print(err, file=sys.stderr)
            _cdr_db.close()
            return False

        else:
            # remove any 0 numbers to keep it from working forever
            _cdr_db.execute(
                f"DELETE FROM {self._TABLE_NAME} WHERE dateTimeOrigination <= 0 OR dateTimeOrigination IS NULL")

        _cdr_db.commit()
        _cdr_db.close()
        return True

    def load_cmr_table(self, cdr_df):
        _cdr_db = sqlite3.connect(self._DB_NAME)

        print("Loading CMR Table")
        try:
            cdr_df.to_sql(self._TABLE_CMR, _cdr_db, if_exists='append', index=False)

        except Exception as err:
            print(err, file=sys.stderr)
            _cdr_db.close()
            return False

        _cdr_db.commit()
        _cdr_db.close()
        return True

    # Simple query returned as List of Tuples
    def query(self, sql, return_type="dict"):
        _cdr_db = sqlite3.connect(self._DB_NAME)

        # if select statement then start cursor
        if sql.lower().startswith("select"):
            try:
                results = pd.read_sql_query(sql, _cdr_db)

            except Exception as err:
                print(f"SQL Error: {err} - {sql}", file=sys.stderr)

            else:
                _cdr_db.commit()
                _cdr_db.close()

                if return_type == "dict":
                    return results.to_dict(orient='records')

                elif return_type == "df":
                    return results

        # if not select then commit the change and close
        else:
            # loop in case its having IO issus writting
            try:
                _cdr_db.execute(sql)

            except Exception as err:
                print(f"SQL Error: {err} - {sql}", file=sys.stderr)
                _cdr_db.close()
                return False

            else:
                _cdr_db.commit()
                _cdr_db.close()
                return True
